#!/usr/bin/env python
"""
File       : reader.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: Module for Writers used in multiple experiment-specific
             workflows.
"""

# system modules
from sys import modules
from time import time

# local modules
from CHAP import Reader


class BinaryFileReader(Reader):
    """Reader for binary files"""
    def _read(self, filename):
        """Return a content of a given file name

        :param filename: name of the binart file to read from
        :return: the content of `filename`
        :rtype: binary
        """

        with open(filename, 'rb') as file:
            data = file.read()
        return data


class MultipleReader(Reader):
    def read(self, readers, **_read_kwargs):
        """Return resuts from multiple `Reader`s.

        :param readers: a dictionary where the keys are specific names
            that are used by the next item in the `Pipeline`, and the
            values are `Reader` configurations.
        :type readers: list[dict]
        :return: The results of calling `Reader.read(**kwargs)` for
            each item configured in `readers`.
        :rtype: list[dict[str,object]]
        """

        t0 = time()
        self.logger.info(f'Executing "read" with {len(readers)} Readers')

        data = []
        for reader_config in readers:
            reader_name = list(reader_config.keys())[0]
            reader_class = getattr(modules[__name__], reader_name)
            reader = reader_class()
            reader_kwargs = reader_config[reader_name]

            # Combine keyword arguments to MultipleReader.read with those to the reader
            # giving precedence to those in the latter
            combined_kwargs = {**_read_kwargs, **reader_kwargs}
            data.extend(reader.read(**combined_kwargs))

        self.logger.info(f'Finished "read" in {time()-t0:.3f} seconds\n')

        return data


class NexusReader(Reader):
    """Reader for NeXus files"""
    def _read(self, filename, nxpath='/'):
        """Return the NeXus object stored at `nxpath` in the nexus
        file `filename`.

        :param filename: name of the NeXus file to read from
        :type filename: str
        :param nxpath: path to a specific loaction in the NeXus file
            to read from, defaults to `'/'`
        :type nxpath: str, optional
        :raises nexusformat.nexus.NeXusError: if `filename` is not a
            NeXus file or `nxpath` is not in `filename`.
        :return: the NeXus structure indicated by `filename` and `nxpath`.
        :rtype: nexusformat.nexus.NXobject
        """

        from nexusformat.nexus import nxload

        nxobject = nxload(filename)[nxpath]
        return nxobject


class URLReader(Reader):
    """Reader for data available over HTTPS"""
    def _read(self, url, headers={}, timeout=10):
        """Make an HTTPS request to the provided URL and return the
        results.  Headers for the request are optional.

        :param url: the URL to read
        :type url: str
        :param headers: headers to attach to the request, defaults to
            `{}`
        :type headers: dict, optional
        :return: the content of the response
        :rtype: object
        """

        import requests

        resp = requests.get(url, headers=headers, timeout=timeout)
        data = resp.content

        self.logger.debug(f'Response content: {data}')

        return data


class YAMLReader(Reader):
    """Reader for YAML files"""
    def _read(self, filename):
        """Return a dictionary from the contents of a yaml file.

        :param filename: name of the YAML file to read from
        :return: the contents of `filename`
        :rtype: dict
        """

        import yaml

        with open(filename) as file:
            data = yaml.safe_load(file)
        return data


if __name__ == '__main__':
    from CHAP.reader import main
    main()



#!/usr/bin/env python
"""
File       : writer.py
Author     : Valentin Kuznetsov <vkuznet AT gmail dot com>
Description: generic Writer module
"""

# system modules
import argparse
from inspect import getfullargspec
import logging
from sys import modules
from time import time


class Writer():
    """Writer represent generic file writer"""

    def __init__(self):
        """Constructor of Writer class"""
        self.__name__ = self.__class__.__name__
        self.logger = logging.getLogger(self.__name__)
        self.logger.propagate = False

    def write(self, data, filename, **_write_kwargs):
        """write API

        :param filename: Name of file to write to
        :param data: data to write to file
        :return: data written to file
        """

        t0 = time()
        self.logger.info(f'Executing "write" with filename={filename}, '
                         f'type(data)={type(data)}, kwargs={_write_kwargs}')

        _valid_write_args = {}
        allowed_args = getfullargspec(self._write).args \
            + getfullargspec(self._write).kwonlyargs
        for k, v in _write_kwargs.items():
            if k in allowed_args:
                _valid_write_args[k] = v
            else:
                self.logger.warning(f'Ignoring invalid arg to _write: {k}')

        data = self._write(data, filename, **_valid_write_args)

        self.logger.info(f'Finished "write" in {time()-t0:.3f} seconds\n')

        return data

    def _write(self, data, filename):
        with open(filename, 'a') as file:
            file.write(data)
        return data


class OptionParser():
    """User based option parser"""
    def __init__(self):
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument(
            '--data', action='store',
            dest='data', default='', help='Input data')
        self.parser.add_argument(
            '--filename', action='store',
            dest='filename', default='', help='Output file')
        self.parser.add_argument(
            '--writer', action='store',
            dest='writer', default='Writer', help='Writer class name')
        self.parser.add_argument(
            '--log-level', choices=logging._nameToLevel.keys(),
            dest='log_level', default='INFO', help='logging level')


def main(opt_parser=OptionParser):
    """Main function"""

    optmgr = opt_parser()
    opts = optmgr.parser.parse_args()
    cls_name = opts.writer
    try:
        writer_cls = getattr(modules[__name__], cls_name)
    except AttributeError:
        print(f'Unsupported writer {cls_name}')
        raise

    writer = writer_cls()
    writer.logger.setLevel(getattr(logging, opts.log_level))
    log_handler = logging.StreamHandler()
    log_handler.setFormatter(logging.Formatter(
        '{name:20}: {message}', style='{'))
    writer.logger.addHandler(log_handler)
    data = writer.write(opts.data, opts.filename)
    print(f'Writer {writer} writes to {opts.filename}, data {data}')


if __name__ == '__main__':
    main()



# third party modules
import numpy as np
from pathlib import PosixPath
from pydantic import (BaseModel,
                      confloat,
                      conint,
                      conlist,
                      constr,
                      FilePath,
                      validator)
from scipy.interpolate import interp1d
from typing import Optional


class MCACeriaCalibrationConfig(BaseModel):
    """
    Class representing metadata required to perform a Ceria calibration for an
    MCA detector.

    :ivar spec_file: Path to the SPEC file containing the CeO2 scan
    :ivar scan_number: Number of the CeO2 scan in `spec_file`
    :ivar scan_step_index: Index of the scan step to use for calibration,
        optional. If not specified, the calibration routine will be performed
        on the average of all MCA spectra for the scan.

    :ivar flux_file: csv file containing station beam energy in eV (column 0)
        and flux (column 1)

    :ivar detector_name: name of the MCA to calibrate
    :ivar num_bins: number of channels on the MCA to calibrate
    :ivar max_energy_kev: maximum channel energy of the MCA in keV

    :ivar hexrd_h5_material_file: path to a HEXRD materials.h5 file containing
        an entry for the material properties.
    :ivar hexrd_h5_material_name: Name of the material entry in
        `hexrd_h5_material_file`, defaults to `'CeO2'`.
    :ivar lattice_parameter_angstrom: lattice spacing in angstrom to use for
        the cubic CeO2 crystal, defaults to `5.41153`.

    :ivar tth_max: detector rotation about hutch x axis, defaults to `90`.
    :ivar hkl_tth_tol: minimum resolvable difference in 2&theta between two
        unique HKL peaks, defaults to `0.15`.

    :ivar fit_include_bin_ranges: list of MCA channel index ranges whose data
        will be included in the calibration routine
    :ivar fit_hkls: list of unique HKL indices to fit peaks for in the
        calibration routine

    :ivar tth_initial_guess: initial guess for 2&theta
    :ivar slope_initial_guess: initial guess for detector channel energy
        correction linear slope, defaults to `1.0`.
    :ivar intercept_initial_guess: initial guess for detector channel energy
        correction y-intercept, defaults to `0.0`.

    :ivar tth_calibrated: calibrated value for 2&theta, defaults to None
    :ivar slope_calibrated: calibrated value for detector channel energy
        correction linear slope, defaults to `None`
    :ivar intercept_calibrated: calibrated value for detector channel energy
        correction y-intercept, defaluts to None

    :ivar max_iter: maximum number of iterations of the calibration routine,
        defaults to `10`.
    :ivar tune_tth_tol: stop iteratively tuning 2&theta when an iteration
        produces a change in the tuned value of 2&theta that is smaller than
        this value, defaults to `1e-8`.
    """

    spec_file: FilePath
    scan_number: conint(gt=0)
    scan_step_index: Optional[conint(ge=0)]

    flux_file: FilePath

    detector_name: constr(strip_whitespace=True, min_length=1)
    num_bins: conint(gt=0)
    max_energy_kev: confloat(gt=0)

    hexrd_h5_material_file: FilePath
    hexrd_h5_material_name: constr(
        strip_whitespace=True, min_length=1) = 'CeO2'
    lattice_parameter_angstrom: confloat(gt=0) = 5.41153

    tth_max: confloat(gt=0, allow_inf_nan=False) = 90.0
    hkl_tth_tol: confloat(gt=0, allow_inf_nan=False) = 0.15

    fit_include_bin_ranges: conlist(min_items=1,
                                    item_type=conlist(item_type=conint(ge=0),
                                                      min_items=2,
                                                      max_items=2))
    fit_hkls: conlist(item_type=conint(ge=0), min_items=1)

    tth_initial_guess: confloat(gt=0, le=tth_max, allow_inf_nan=False)
    slope_initial_guess: float = 1.0
    intercept_initial_guess: float = 0.0
    tth_calibrated: Optional[confloat(gt=0, allow_inf_nan=False)]
    slope_calibrated: Optional[confloat(allow_inf_nan=False)]
    intercept_calibrated: Optional[confloat(allow_inf_nan=False)]

    max_iter: conint(gt=0) = 10
    tune_tth_tol: confloat(ge=0) = 1e-8

    @validator('fit_include_bin_ranges', each_item=True)
    def validate_include_bin_range(cls, value, values):
        """Ensure no bin ranges are outside the boundary of the detector"""

        num_bins = values.get('num_bins')
        value[1] = min(value[1], num_bins)
        return value

    def mca_data(self):
        """Get the 1D array of MCA data to use for calibration.

        :return: MCA data
        :rtype: np.ndarray
        """
        # local modules
        from CHAP.common.utils.scanparsers \
            import SMBMCAScanParser as ScanParser

        scanparser = ScanParser(self.spec_file, self.scan_number)
        if self.scan_step_index is None:
            data = scanparser.get_all_detector_data(self.detector_name)
            if scanparser.spec_scan_npts > 1:
                data = np.average(data, axis=1)
            else:
                data = data[0]
        else:
            data = scanparser.get_detector_data(
                self.detector_name, self.scan_step_index)

        return np.array(data)

    def mca_mask(self):
        """Get a boolean mask array to use on MCA data before fitting.

        :return: boolean mask array
        :rtype: numpy.ndarray
        """

        mask = np.asarray([False]*self.num_bins)
        bin_indices = np.arange(self.num_bins)
        for min_, max_ in self.fit_include_bin_ranges:
            _mask = np.logical_and(bin_indices > min_, bin_indices < max_)
            mask = np.logical_or(mask, _mask)

        return mask

    def flux_correction_interpolation_function(self):
        """
        Get an interpolation function to correct MCA data for relative energy
        flux of the incident beam.

        :return: energy flux correction interpolation function
        :rtype: scipy.interpolate._polyint._Interpolator1D
        """

        flux = np.loadtxt(self.flux_file)
        energies = flux[:,0]/1.e3
        relative_intensities = flux[:,1]/np.max(flux[:,1])
        interpolation_function = interp1d(energies, relative_intensities)
        return interpolation_function

    def material(self):
        """Get CeO2 as a `CHAP.common.utils.material.Material` object.

        :return: CeO2 material
        :rtype: CHAP.common.utils.material.Material
        """
        # local modules
        from CHAP.common.utils.material import Material

        material = Material(
            material_name=self.hexrd_h5_material_name,
            material_file=self.hexrd_h5_material_file,
            lattice_parameters_angstroms=self.lattice_parameter_angstrom)
        # The following kwargs will be needed if we allow the material to be
        # built using xrayutilities (for now, we only allow hexrd to make the
        # material):
        #   sgnum=225,
        #   atoms=['Ce4p', 'O2mdot'],
        #   pos=[(0.,0.,0.), (0.25,0.75,0.75)],
        #   enrgy=50000.)
        # Why do we need to specify an energy to get HKLs when using
        # xrayutilities?
        return material

    def unique_ds(self):
        """Get a list of unique HKLs and their lattice spacings

        :return: unique HKLs and their lattice spacings in angstroms
        :rtype: np.ndarray, np.ndarray
        """

        unique_hkls, unique_ds = self.material().get_ds_unique(
            tth_tol=self.hkl_tth_tol, tth_max=self.tth_max)

        return unique_hkls, unique_ds

    def fit_ds(self):
        """
        Get a list of HKLs and their lattice spacings that will be fit in the
        calibration routine

        :return: HKLs to fit and their lattice spacings in angstroms
        :rtype: np.ndarray, np.ndarray
        """

        unique_hkls, unique_ds = self.unique_ds()

        fit_hkls = np.array([unique_hkls[i] for i in self.fit_hkls])
        fit_ds = np.array([unique_ds[i] for i in self.fit_hkls])

        return fit_hkls, fit_ds

    def dict(self):
        """
        Return a representation of this configuration in a dictionary that is
        suitable for dumping to a YAML file (one that converts all instances of
        fields with type `PosixPath` to `str`).

        :return: dictionary representation of the configuration.
        :rtype: dict
        """

        d = super().dict()
        for k,v in d.items():
            if isinstance(v, PosixPath):
                d[k] = str(v)
        return d
